import argparse
import asyncio
import os
import sys

from dotenv import dotenv_values, find_dotenv
from mcp import ClientSession, StdioServerParameters, stdio_client
from mcp.types import TextContent

from .main import __version__, app_factory


def main():
    name_version = f'Logfire MCP v{__version__}'
    parser = argparse.ArgumentParser(
        prog='logfire-mcp',
        description=f'{name_version}\n\nSee github.com/pydantic/logfire-mcp',
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument(
        '--read-token',
        type=str,
        help='Pydantic Logfire read token. Can also be set via LOGFIRE_READ_TOKEN environment variable.',
    )
    parser.add_argument(
        '--base-url',
        type=str,
        required=False,
        help='Pydantic Logfire base URL. Can also be set via LOGFIRE_BASE_URL environment variable.',
    )
    parser.add_argument('--test', action='store_true', help='Test the MCP server and exit')
    parser.add_argument('--version', action='store_true', help='Show version and exit')
    args = parser.parse_args()
    if args.version:
        print(name_version)
        return

    # Get token from args or environment
    logfire_read_token, source = get_read_token(args)
    if not logfire_read_token:
        parser.error(
            'Pydantic Logfire read token must be provided either via --read-token argument '
            'or LOGFIRE_READ_TOKEN environment variable'
        )

    logfire_base_url = args.base_url or os.getenv('LOGFIRE_BASE_URL')
    if args.test:
        asyncio.run(test(logfire_read_token, logfire_base_url, source))
    else:
        app = app_factory(logfire_read_token, logfire_base_url)
        app.run(transport='stdio')


async def test(logfire_read_token: str, logfire_base_url: str | None, source: str):
    print('testing Logfire MCP server:\n')
    print(f'logfire_read_token: `{logfire_read_token[:12]}...{logfire_read_token[-5:]}` from {source}\n')

    args = ['-m', 'logfire_mcp', '--read-token', logfire_read_token]
    if logfire_base_url:
        print(f'logfire_base_url: `{logfire_base_url}`')
        args += ['--base-url', logfire_base_url]

    server_params = StdioServerParameters(command=sys.executable, args=args)
    async with stdio_client(server_params) as (read, write):
        async with ClientSession(read, write) as session:
            await session.initialize()

            tools = await session.list_tools()
            print('tools:')
            for tool in tools.tools:
                print(f'  - {tool.name}')

            list_resources = await session.list_resources()
            print('resources:')
            for resource in list_resources.resources:
                print(f'  - {resource.name}')

            for tool in 'sql_reference', 'get_logfire_records_schema':
                print(f'\ncalling `{tool}`:')
                output = await session.call_tool(tool)
                # debug(output)
                content = output.content[0]
                assert isinstance(content, TextContent), f'Expected TextContent, got {type(content)}'
                if len(content.text) < 200:
                    print(f'> {content.text.strip()}')
                else:
                    first_line = content.text.strip().split('\n', 1)[0]
                    print(f'> {first_line}... ({len(content.text) - len(first_line)} more characters)\n')


def get_read_token(args: argparse.Namespace) -> tuple[str | None, str]:
    if args.read_token:
        return args.read_token, 'CLI argument'
    elif token := os.getenv('LOGFIRE_READ_TOKEN'):
        return token, 'environment variable'
    else:
        return dotenv_values(dotenv_path=find_dotenv(usecwd=True)).get('LOGFIRE_READ_TOKEN'), 'dotenv file'


if __name__ == '__main__':
    main()



from pathlib import Path
from typing import TypedDict

import pytest
from jinja2 import Environment, FileSystemLoader
from mcp.client.session import ClientSession

env = Environment(loader=FileSystemLoader(Path(__file__).parent))
template = env.get_template('README.md.jinja')


pytestmark = [pytest.mark.vcr, pytest.mark.anyio]


class Argument(TypedDict):
    name: str
    description: str
    type: str


class Tool(TypedDict):
    name: str
    description: str
    arguments: list[Argument]


async def test_generate_readme(session: ClientSession) -> None:
    tools: list[Tool] = []
    mcp_tools = await session.list_tools()

    for tool in mcp_tools.tools:
        assert tool.description
        description = tool.description.split('\n', 1)[0].strip()

        arguments: list[Argument] = []
        for argument_name, argument_schema in tool.inputSchema['properties'].items():
            arguments.append(
                {'name': argument_name, 'description': argument_schema['description'], 'type': argument_schema['type']}
            )
        tools.append({'name': tool.name, 'description': description, 'arguments': arguments})

    readme = template.render(tools=tools)
    with open('README.md', 'w') as f:
        f.write(readme)



import os
from collections.abc import AsyncGenerator

import pytest
from mcp.client.session import ClientSession
from mcp.server.fastmcp import FastMCP
from mcp.shared.memory import create_connected_server_and_client_session

from logfire_mcp.__main__ import app_factory


@pytest.fixture
def anyio_backend():
    return 'asyncio'


@pytest.fixture
def vcr_config():
    return {'filter_headers': [('authorization', None)]}


@pytest.fixture
async def logfire_read_token() -> str:
    # To get a read token, go to https://logfire-us.pydantic.dev/kludex/logfire-mcp/settings/read-tokens/.
    return os.getenv('LOGFIRE_READ_TOKEN', 'fake-token')


@pytest.fixture
def app(logfire_read_token: str) -> FastMCP:
    return app_factory(logfire_read_token)


@pytest.fixture
async def session(app: FastMCP) -> AsyncGenerator[ClientSession]:
    mcp_server = app._mcp_server  # type: ignore
    async with create_connected_server_and_client_session(mcp_server, raise_exceptions=True) as _session:
        yield _session



import pytest
from mcp.client.session import ClientSession
from mcp.types import TextContent

pytestmark = [pytest.mark.vcr, pytest.mark.anyio]


async def test_logfire_link(session: ClientSession) -> None:
    result = await session.call_tool('logfire_link', {'trace_id': '019837e6ba8ab0ede383b398b6706f28'})

    assert result.content == [
        TextContent(
            type='text',
            text='https://logfire-us.pydantic.dev/logfire/gateway?q=trace_id%3D%27019837e6ba8ab0ede383b398b6706f28%27',
        )
    ]



import pytest
from inline_snapshot import snapshot
from mcp.client.session import ClientSession
from mcp.types import TextContent

pytestmark = [pytest.mark.vcr, pytest.mark.anyio]


async def test_schema_reference(session: ClientSession) -> None:
    result = await session.call_tool('schema_reference')

    assert result.content == snapshot(
        [
            TextContent(
                type='text',
                text="""\
CREATE TABLE records (
    attributes TEXT,
    attributes_json_schema TEXT,
    created_at TIMESTAMP WITH TIME ZONE NOT NULL,
    day DATE NOT NULL,
    deployment_environment TEXT,
    duration DOUBLE PRECISION,
    end_timestamp TIMESTAMP WITH TIME ZONE NOT NULL,
    exception_message TEXT,
    exception_stacktrace TEXT,
    exception_type TEXT,
    http_method TEXT,
    http_response_status_code INTEGER,
    http_route TEXT,
    is_exception BOOLEAN,
    kind TEXT NOT NULL,
    level INTEGER NOT NULL,
    log_body TEXT,
    message TEXT NOT NULL,
    otel_events TEXT,
    otel_links TEXT,
    otel_resource_attributes TEXT,
    otel_scope_attributes TEXT,
    otel_scope_name TEXT,
    otel_scope_version TEXT,
    otel_status_code TEXT,
    otel_status_message TEXT,
    parent_span_id TEXT,
    process_pid INTEGER,
    project_id TEXT NOT NULL,
    service_instance_id TEXT,
    service_name TEXT NOT NULL,
    service_namespace TEXT,
    service_version TEXT,
    span_id TEXT NOT NULL,
    span_name TEXT NOT NULL,
    start_timestamp TIMESTAMP WITH TIME ZONE NOT NULL,
    tags TEXT[],
    telemetry_sdk_language TEXT,
    telemetry_sdk_name TEXT,
    telemetry_sdk_version TEXT,
    trace_id TEXT NOT NULL,
    url_full TEXT,
    url_path TEXT,
    url_query TEXT
);

CREATE TABLE metrics (
    aggregation_temporality TEXT,
    attributes TEXT,
    attributes_json_schema TEXT,
    created_at TIMESTAMP WITH TIME ZONE NOT NULL,
    day DATE NOT NULL,
    deployment_environment TEXT,
    exemplars TEXT,
    exp_histogram_negative_bucket_counts INTEGER[],
    exp_histogram_negative_bucket_counts_offset INTEGER,
    exp_histogram_positive_bucket_counts INTEGER[],
    exp_histogram_positive_bucket_counts_offset INTEGER,
    exp_histogram_scale INTEGER,
    exp_histogram_zero_count INTEGER,
    exp_histogram_zero_threshold DOUBLE PRECISION,
    histogram_bucket_counts INTEGER[],
    histogram_count INTEGER,
    histogram_explicit_bounds DOUBLE PRECISION[],
    histogram_max DOUBLE PRECISION,
    histogram_min DOUBLE PRECISION,
    histogram_sum DOUBLE PRECISION,
    is_monotonic BOOLEAN,
    metric_description TEXT,
    metric_name TEXT NOT NULL,
    metric_type TEXT NOT NULL,
    otel_resource_attributes TEXT,
    otel_scope_attributes TEXT,
    otel_scope_name TEXT,
    otel_scope_version TEXT,
    process_pid INTEGER,
    project_id TEXT NOT NULL,
    recorded_timestamp TIMESTAMP WITH TIME ZONE,
    scalar_value DOUBLE PRECISION,
    service_instance_id TEXT,
    service_name TEXT NOT NULL,
    service_namespace TEXT,
    service_version TEXT,
    start_timestamp TIMESTAMP WITH TIME ZONE,
    telemetry_sdk_language TEXT,
    telemetry_sdk_name TEXT,
    telemetry_sdk_version TEXT,
    unit TEXT NOT NULL
);\
""",
            )
        ]
    )



from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from dataclasses import dataclass
from datetime import UTC, datetime, timedelta
from importlib.metadata import version
from typing import Annotated, Any, TypedDict, cast

from logfire.experimental.query_client import AsyncLogfireQueryClient
from mcp.server.fastmcp import Context, FastMCP
from mcp.server.session import ServerSession
from pydantic import Field, WithJsonSchema


@dataclass
class MCPState:
    logfire_client: AsyncLogfireQueryClient


HOUR = 60  # minutes
DAY = 24 * HOUR

__version__ = version('logfire-mcp')

Age = Annotated[
    int,
    Field(
        ge=0,
        le=7 * DAY,
        description='Number of minutes to look back, e.g. 30 for last 30 minutes. Maximum allowed value is 7 days.',
    ),
    WithJsonSchema({'type': 'integer'}),
]


async def find_exceptions_in_file(
    ctx: Context[ServerSession, MCPState],
    filepath: Annotated[str, Field(description='The path to the file to find exceptions in.')],
    age: Age,
) -> list[Any]:
    """Get the details about the 10 most recent exceptions on the file."""
    logfire_client = ctx.request_context.lifespan_context.logfire_client
    min_timestamp = datetime.now(UTC) - timedelta(minutes=age)
    result = await logfire_client.query_json_rows(
        f"""\
        SELECT
            created_at,
            message,
            exception_type,
            exception_message,
            exception_stacktrace
        FROM records
        WHERE is_exception = true
            AND exception_stacktrace like '%{filepath}%'
        ORDER BY created_at DESC
        LIMIT 10
    """,
        min_timestamp=min_timestamp,
    )
    return result['rows']


async def arbitrary_query(
    ctx: Context[ServerSession, MCPState],
    query: Annotated[str, Field(description='The query to run, as a SQL string.')],
    age: Age,
) -> list[Any]:
    """Run an arbitrary query on the Pydantic Logfire database.

    The SQL reference is available via the `sql_reference` tool.
    """
    logfire_client = ctx.request_context.lifespan_context.logfire_client
    min_timestamp = datetime.now(UTC) - timedelta(minutes=age)
    result = await logfire_client.query_json_rows(query, min_timestamp=min_timestamp)
    return result['rows']


async def schema_reference(ctx: Context[ServerSession, MCPState]) -> str:
    """The database schema for the Logfire DataFusion database.

    This includes all tables, columns, and their types as well as descriptions.
    For example:

    ```sql
    -- The records table contains spans and logs.
    CREATE TABLE records (
        message TEXT, -- The message of the record
        span_name TEXT, -- The name of the span, message is usually templated from this
        trace_id TEXT, -- The trace ID, identifies a group of spans in a trace
        exception_type TEXT, -- The type of the exception
        exception_message TEXT, -- The message of the exception
        -- other columns...
    );
    ```
    The SQL syntax is similar to Postgres, although the query engine is actually Apache DataFusion.

    To access nested JSON fields e.g. in the `attributes` column use the `->` and `->>` operators.
    You may need to cast the result of these operators e.g. `(attributes->'cost')::float + 10`.

    You should apply as much filtering as reasonable to reduce the amount of data queried.
    Filters on `start_timestamp`, `service_name`, `span_name`, `metric_name`, `trace_id` are efficient.
    """
    logfire_client = ctx.request_context.lifespan_context.logfire_client
    response = await logfire_client.client.get('/v1/schemas')
    schema_data = response.json()

    def schema_to_sql(schema_json: dict[str, Any]) -> str:
        sql_commands: list[str] = []
        for table in schema_json.get('tables', []):
            table_name = table['name']
            columns: list[str] = []

            for col_name, col_info in table['schema'].items():
                data_type = col_info['data_type']
                nullable = col_info.get('nullable', True)
                description = col_info.get('description', '').strip()

                column_def = f'{col_name} {data_type}'
                if not nullable:
                    column_def += ' NOT NULL'
                if description:
                    column_def += f' -- {description}'

                columns.append(column_def)

            create_table = f'CREATE TABLE {table_name} (\n    ' + ',\n    '.join(columns) + '\n);'
            sql_commands.append(create_table)

        return '\n\n'.join(sql_commands)

    return schema_to_sql(schema_data)


async def logfire_link(
    ctx: Context[ServerSession, MCPState],
    trace_id: Annotated[str, Field(description='The trace ID to link to.')],
) -> str:
    """Creates a link to help the user to view the trace in the Logfire UI."""
    logfire_client = ctx.request_context.lifespan_context.logfire_client
    response = await logfire_client.client.get('/v1/read-token-info')
    read_token_info = cast(ReadTokenInfo, response.json())
    organization_name = read_token_info['organization_name']
    project_name = read_token_info['project_name']

    url = logfire_client.client.base_url
    url = url.join(f'{organization_name}/{project_name}')
    url = url.copy_add_param('q', f"trace_id='{trace_id}'")
    return str(url)


def app_factory(logfire_read_token: str, logfire_base_url: str | None = None) -> FastMCP:
    @asynccontextmanager
    async def lifespan(server: FastMCP) -> AsyncIterator[MCPState]:
        # print to stderr so we this message doesn't get read by the MCP client
        headers = {'User-Agent': f'logfire-mcp/{__version__}'}
        async with AsyncLogfireQueryClient(logfire_read_token, headers=headers, base_url=logfire_base_url) as client:
            yield MCPState(logfire_client=client)

    mcp = FastMCP('Logfire', lifespan=lifespan)
    mcp.tool()(find_exceptions_in_file)
    mcp.tool()(arbitrary_query)
    mcp.tool()(logfire_link)
    mcp.tool()(schema_reference)

    return mcp


class ReadTokenInfo(TypedDict):
    token_id: str
    organization_id: str
    project_id: str
    organization_name: str
    project_name: str
